######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Module for data store class. """
from dataclasses import dataclass
import os
from shutil import copyfile
from PySide6.QtCore import Slot
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QApplication, QFileDialog, QMenu
from spine_engine.project_item.project_item_resource import ProjectItemResource, database_resource
from spine_engine.utils.serialization import deserialize_path, serialize_path
from spinedb_api.helpers import remove_credentials_from_url, vacuum
from spinetoolbox.helpers import create_dir
from spinetoolbox.project_item.project_item import ProjectItem
from spinetoolbox.spine_db_editor.editors import db_editor_registry
from spinetoolbox.spine_db_editor.widgets.multi_spine_db_editor import open_db_editor
from spinetoolbox.widgets.custom_qwidgets import SelectDatabaseItemsDialog
from ..database_validation import DatabaseConnectionValidator
from ..utils import convert_to_sqlalchemy_url, database_label
from .commands import UpdateDSURLCommand
from .executable_item import ExecutableItem
from .item_info import ItemInfo
from .output_resources import scan_for_resources


@dataclass(frozen=True)
class ReplaceableResource:
    resource: ProjectItemResource
    is_valid: bool


class DataStore(ProjectItem):
    def __init__(self, name, description, x, y, toolbox, project, url):
        """Data Store class.

        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): the project this item belongs to
            url (dict, optional): SQLAlchemy url
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self.logs_dir = os.path.join(self.data_dir, "logs")
        try:
            create_dir(self.logs_dir)
        except OSError:
            self._logger.msg_error.emit(f"[OSError] Creating directory {self.logs_dir} failed. Check permissions.")
        if url is None:
            url = {"dialect": "sqlite"}
        self._url = self.parse_url(url)
        self._url_validated = False
        self._resource_to_replace = None
        self._multi_db_editors_open = {}
        self._open_url_menu = QMenu("Open URL in Spine DB editor", self._toolbox)
        self._open_url_action = QAction("Open URL in Spine DB editor")
        self._open_url_menu.triggered.connect(self._handle_open_url_menu_triggered)
        self._open_url_action.triggered.connect(self.open_url_in_spine_db_editor)
        self._purge_settings = None
        self._purge_dialog = None
        self._database_validator = DatabaseConnectionValidator(self)
        self._toolbox.db_mngr.database_clean_changed.connect(self._set_database_clean)

    def get_db_map(self):
        """Returns the db map for the Data Store.

        Returns:
            DatabaseMapping: database mapping or None if URL/database is invalid
        """
        sa_url = self.sql_alchemy_url()
        if sa_url is None:
            return None
        return self._toolbox.db_mngr.get_db_map(sa_url, self._logger)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def set_up(self):
        """See base class."""
        super().set_up()
        self._actions.clear()
        self._actions.append(self._open_url_action)

    def parse_url(self, url):
        """Return a complete url dictionary from the given dict or string"""
        base_url = {"dialect": "", "username": "", "password": "", "host": "", "port": "", "database": "", "schema": ""}
        if isinstance(url, dict):
            if url.get("dialect") == "sqlite" and (database := url.get("database")):
                # Convert relative database path back to absolute
                url["database"] = os.path.abspath(os.path.join(self._project.project_dir, database))
            for key, value in url.items():
                if value is not None:
                    base_url[key] = value
        return base_url

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.url_selector_widget.url_changed] = self._update_url_from_properties
        s[self._properties_ui.pushButton_ds_open_editor.clicked] = self.open_url_in_spine_db_editor
        s[self._properties_ui.pushButton_create_new_spine_db.clicked] = self.create_new_spine_database
        s[self._properties_ui.toolButton_copy_url.clicked] = self.copy_url
        s[self._properties_ui.toolButton_vacuum.clicked] = self.vacuum
        s[self._properties_ui.purge_button.clicked] = self._show_purge_dialog
        return s

    def restore_selections(self):
        """Load url into selections."""
        self.load_url_into_selections(self._url)

    def url(self):
        """Returns the url attribute."""
        return self._url

    def sql_alchemy_url(self):
        """Returns the URL as an SQLAlchemy URL object or None if no URL is set."""
        return convert_to_sqlalchemy_url(self._url, self.name, self._logger)

    def select_sqlite_file(self):
        """Open file browser where user can select the path to an SQLite file that they want to use."""
        candidate_path = os.path.abspath(self._url["database"]) if self._url["database"] else self.data_dir
        answer = QFileDialog.getOpenFileName(self._toolbox, "Select SQLite file", candidate_path)
        file_path = answer[0]
        return file_path if file_path else None

    def _new_sqlite_file(self):
        """Shows a file dialog and creates a new sqlite file at the chosen path.

        Returns:
            bool: True if the file was created successfully, False otherwise
        """
        candidate_path = os.path.abspath(os.path.join(self.data_dir, self.name + ".sqlite"))
        answer = QFileDialog.getSaveFileName(self._toolbox, "Create SQLite file", candidate_path)
        file_path = answer[0]
        if not file_path:  # Cancel button clicked
            return False
        abs_path = os.path.abspath(file_path)
        url = dict(self._url)
        url["database"] = abs_path
        sa_url = convert_to_sqlalchemy_url(url, self.name)
        self._toolbox.db_mngr.create_new_spine_database(sa_url, self._logger, overwrite=True)
        self.update_url(dialect="sqlite", database=abs_path)
        return True

    def load_url_into_selections(self, url):
        """Load given url attribute into shared widget selections.

        Args:
            url (dict): URL dict
        """
        if not self._active or not url:
            return
        self._properties_ui.url_selector_widget.set_url(url)
        self._update_actions_enabled()

    @Slot()
    def _update_url_from_properties(self):
        """Gets new URL from Properties tab."""
        url = self._properties_ui.url_selector_widget.url_dict()
        self.update_url(**url)

    def update_url(self, **kwargs):
        """Pushes a command to update the URL to the undo stack.

        Args:
            **kwargs: URL keys and values
        """
        invalidating_url = self._url_validated
        kwargs = {k: v for k, v in kwargs.items() if v != self._url[k]}
        if not kwargs:
            return False
        self._toolbox.undo_stack.push(UpdateDSURLCommand(self.name, invalidating_url, self._project, **kwargs))
        return True

    def do_update_url(self, **kwargs):
        """Updates URL.

        Args:
            **kwargs: URL keys and values
        """
        was_valid = self._url_validated
        self._url_validated = False
        old_url = convert_to_sqlalchemy_url(self._url, self.name)
        if old_url is not None and was_valid:
            self._toolbox.db_mngr.name_registry.unregister(old_url, self.name)
        new_dialect = kwargs.get("dialect")
        if new_dialect == "sqlite":
            kwargs.update({"username": "", "password": "", "host": "", "port": "", "schema": ""})
        self._url.update(kwargs)
        new_url = convert_to_sqlalchemy_url(self._url, self.name)
        self.load_url_into_selections(self._url)
        if old_url and new_url:
            old_resource = database_resource(self.name, str(old_url), label=database_label(self.name), filterable=True)
            self._resource_to_replace = ReplaceableResource(old_resource, was_valid)
        else:
            self._resources_to_predecessors_changed()
            self._resources_to_successors_changed()
        self._check_notifications()

    def has_listeners(self):
        """Checks whether the Data Store has listeners or not

        Returns:
             (bool): True if there are listeners for the Data Store, False otherwise
        """
        if self._multi_db_editors_open:
            return bool(self._toolbox.db_mngr.db_map_listeners(self.get_db_map()))
        return False

    def _update_actions_enabled(self):
        url_exists = convert_to_sqlalchemy_url(self._url, self.name) is not None
        url_valid = url_exists and self._url_validated
        self._open_url_action.setEnabled(url_valid)
        self._open_url_menu.setEnabled(url_valid)
        if not self._active:
            return
        is_sqlite = self._url["dialect"].lower() == "sqlite"
        self._properties_ui.pushButton_ds_open_editor.setEnabled(url_valid)
        self._properties_ui.pushButton_create_new_spine_db.setEnabled(is_sqlite or url_valid)
        self._properties_ui.purge_button.setEnabled(url_valid)
        self._properties_ui.toolButton_copy_url.setEnabled(url_exists)
        self._properties_ui.toolButton_vacuum.setEnabled(is_sqlite and url_valid)

    @Slot(bool)
    def _show_purge_dialog(self, _=False):
        """Shows the Purge dialog."""
        if self._purge_dialog is not None:
            self._purge_dialog.raise_()
            return
        self._purge_dialog = SelectDatabaseItemsDialog(self._purge_settings, "Purge", self._toolbox)
        self._purge_dialog.setWindowTitle(f"Purge items from {self.name}")
        self._purge_dialog.accepted.connect(self._purge)
        self._purge_dialog.destroyed.connect(self._clean_up_purge_dialog)
        self._purge_dialog.show()

    @Slot()
    def _purge(self):
        """Purges the database."""
        self._purge_settings = self._purge_dialog.get_checked_states()
        db_map = self.get_db_map()
        if db_map is None:
            return
        db_map_purge_data = {db_map: {item_type for item_type, checked in self._purge_settings.items() if checked}}
        self._toolbox.db_mngr.purge_items(db_map_purge_data)
        self._toolbox.db_mngr.commit_session("Purge the database.", db_map)
        self._logger.msg_success.emit(f"<b>{self.name}</b>: Database purged.")

    @Slot()
    def _clean_up_purge_dialog(self):
        """Cleans up stuff after Purge dialog has been closed."""
        self._purge_dialog = None

    def actions(self):
        self._multi_db_editors_open = {x.name(): x for x in db_editor_registry.windows()}
        if not self._multi_db_editors_open:
            return super().actions()
        self._open_url_menu.clear()
        self._open_url_menu.addAction("New window")
        self._open_url_menu.addSeparator()
        for name in self._multi_db_editors_open:
            self._open_url_menu.addAction(name)
        return [self._open_url_menu.menuAction()]

    @Slot(QAction)
    def _handle_open_url_menu_triggered(self, action):
        """Opens current url."""
        if action.text() == "New window":
            self._open_url_in_new_db_editor()
            return
        db_editor = self._multi_db_editors_open[action.text()]
        self._open_url_in_existing_db_editor(db_editor)

    @Slot(bool)
    def open_url_in_spine_db_editor(self, checked=False):
        """Opens current url in the Spine database editor."""
        self._open_spine_db_editor(reuse_existing=True)

    def _open_url_in_new_db_editor(self, checked=False):
        self._open_spine_db_editor(reuse_existing=False)

    def _open_spine_db_editor(self, reuse_existing):
        """Opens Data Store's URL in Spine Database editor.

        Args:
            reuse_existing (bool): if True and the URL is already open, just raise the window
        """
        if not self._url_validated:
            self._logger.msg_error.emit(
                f"<b>{self.name}</b> is still validating the database URL or the URL is invalid."
            )
            return
        sa_url = self.sql_alchemy_url()
        if sa_url is not None:
            open_db_editor([sa_url], self._toolbox.db_mngr, reuse_existing)
        self._check_notifications()

    def _open_url_in_existing_db_editor(self, db_editor):
        if not self._url_validated:
            db_editor.add_message(f"<b>{self.name}</b> is still validating the database URL or the URL is invalid.")
            return
        sa_url = self.sql_alchemy_url()
        if sa_url is not None:
            db_editor.add_new_tab([sa_url])
            db_editor.raise_()
            db_editor.activateWindow()

    @Slot(bool)
    def copy_url(self, checked=False):
        """Copy db url to clipboard."""
        sa_url = convert_to_sqlalchemy_url(self._url, self.name, self._logger)
        if sa_url is None:
            return
        sa_url = sa_url.set(password=None)
        QApplication.clipboard().setText(str(sa_url))
        self._logger.msg.emit(f"Database url <b>{sa_url}</b> copied to clipboard")

    @Slot(bool)
    def vacuum(self, checked=False):
        sa_url = convert_to_sqlalchemy_url(self._url, self.name, self._logger)
        freed, unit = vacuum(sa_url)
        self._logger.msg.emit(f"Vacuum finished, {freed} {unit} freed from {sa_url.database}")

    @Slot(bool)
    def create_new_spine_database(self, checked=False):
        """Create new (empty) Spine database."""
        if self._url["dialect"] == "sqlite":
            sqlite_file_creation_successful = self._new_sqlite_file()
            if not sqlite_file_creation_successful:  # User cancelled
                return
        elif self._url["dialect"] == "mysql":
            sa_url = convert_to_sqlalchemy_url(self._url, self.name, self._logger)
            if sa_url:
                self._toolbox.db_mngr.create_new_spine_database(sa_url, self._logger)
        else:
            self._logger.msg_error.emit(f"Unknown data store dialect: {self._url['dialect']}.")
        self._check_notifications()

    def _check_notifications(self):
        """Updates the SqlAlchemy format URL and checks for notifications"""
        self.clear_notifications()
        self._update_actions_enabled()
        sa_url = convert_to_sqlalchemy_url(self._url, self.name)
        if sa_url is None:
            self.clear_notifications()
            self.add_notification(
                "The URL for this Data Store is not correctly set. Set it in the Data Store Properties panel."
            )
            return
        self._database_validator.validate_url(
            self._url["dialect"], sa_url, self._set_invalid_url_notification, self._accept_url
        )

    @Slot(object, bool)
    def _set_database_clean(self, db_map, clean):
        """Updates database cleanliness status.

        Args:
            db_map (DatabaseMapping): database mapping
            clean (bool): True if database is clean, False otherwise
        """
        if self._url_validated and db_map is not self.get_db_map():
            return
        self._notify_about_dirtiness(clean)

    def _notify_about_dirtiness(self, clean):
        """
        Handles the notification for the dirtiness of the Data Store

        Args:
            clean (bool): Whether the db_map corresponding to the DS is clean
        """
        if not clean:
            self.add_notification(f"{self.name} has uncommitted changes")
        else:
            self.remove_notification(f"{self.name} has uncommitted changes")

    @Slot(str, object)
    def _set_invalid_url_notification(self, error_message, url):
        """Sets a single notification that warns about broken URL.

        Args:
            error_message (str): URL failure message
            url (URL): SqlAlchemy URL
        """
        self.clear_notifications()
        self.add_notification(
            f"Couldn't connect to the database <b>{remove_credentials_from_url(str(url))}</b>: {error_message}"
        )
        if self._resource_to_replace is None:
            self._resources_to_predecessors_changed()
            self._resources_to_successors_changed()

    @Slot(object)
    def _accept_url(self, url):
        """Sets URL as validated and updates advertised resources."""
        db_map = self.get_db_map()
        if db_map:
            self._toolbox.db_mngr.name_registry.register(db_map.sa_url, self.name)
            clean = not self._toolbox.db_mngr.is_dirty(db_map)
            self._notify_about_dirtiness(clean)
        else:
            self.add_notification(f"Database is not valid.")
            return
        self._url_validated = True
        self.clear_other_notifications(f"{self.name} has uncommitted changes")
        if self._resource_to_replace is not None and self._resource_to_replace.is_valid:
            old = self._resource_to_replace.resource
            sa_url = convert_to_sqlalchemy_url(self._url, self.name)
            new = database_resource(self.name, str(sa_url), label=database_label(self.name), filterable=True)
            self._resources_to_predecessors_replaced([old], [new])
            self._resources_to_successors_replaced([old], [new])
            self._resource_to_replace = None
        else:
            self._resources_to_predecessors_changed()
            self._resources_to_successors_changed()
        self._update_actions_enabled()

    def is_url_validated(self):
        """Tests whether the URL has been validated.

        Returns:
            bool: True if URL has been validated, False otherwise
        """
        return self._url_validated

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        d["url"] = dict(self.url())
        # If database key is a file, change the path to relative
        if d["url"]["dialect"] == "sqlite" and d["url"]["database"]:
            d["url"]["database"] = serialize_path(d["url"]["database"], self._project.project_dir)
        return d

    @staticmethod
    def item_dict_local_entries():
        """See base class."""
        return [("url", "username"), ("url", "password")]

    def copy_local_data(self, item_dict):
        """See base class."""
        original_data_dir = item_dict.get("original_data_dir")  # (str) original dir of duplicated ProjectItem
        original_db_url = item_dict.get("original_db_url")  # (dict) original url of the duplicated ProjectItem
        duplicate_files = item_dict.get("duplicate_files")  # (bool) Flag indicating if linked files should be copied
        if original_data_dir is None and original_db_url is None and duplicate_files is None:
            return
        original_project_dir = original_data_dir.split(".spinetoolbox")[0]
        linked_db_dialect = original_db_url.get("dialect")
        if linked_db_dialect != "sqlite":
            return
        original_db = original_db_url.get("database")
        if isinstance(original_db, dict):
            if original_db["relative"]:
                linked_db_path = os.path.join(original_project_dir, original_db["path"])
            else:
                linked_db_path = original_db["path"]
        else:
            linked_db_path = original_db
        if duplicate_files:
            db_copy_destination = os.path.join(self.data_dir, os.path.basename(linked_db_path))
            copyfile(linked_db_path, db_copy_destination)
            self._url["database"] = db_copy_destination
        else:
            self._url["database"] = linked_db_path

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        url = item_dict["url"]
        if url and not isinstance(url["database"], str):
            url["database"] = deserialize_path(url["database"], project.project_dir)
        return DataStore(name, description, x, y, toolbox, project, url)

    def rename(self, new_name, rename_data_dir_message):
        """See base class."""
        old_data_dir = os.path.abspath(self.data_dir)
        old_name = self.name
        if not super().rename(new_name, rename_data_dir_message):
            return False
        if self._url_validated:
            sa_url = self.sql_alchemy_url()
            self._toolbox.db_mngr.name_registry.unregister(sa_url, old_name)
        # If dialect is sqlite and db line edit refers to a file in the old data_dir, db line edit needs updating
        if self._url["dialect"] == "sqlite":
            db_dir, db_filename = os.path.split(os.path.abspath(self._url["database"].strip()))
            if os.path.normcase(db_dir) == os.path.normcase(old_data_dir):
                database = os.path.join(self.data_dir, db_filename)  # NOTE: data_dir has been updated at this point
                # Check that the db was moved successfully to the new data_dir
                if os.path.exists(database):
                    self._url.update(database=database)
                    self.load_url_into_selections(self._url)
            self._check_notifications()
        return True

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Importer":
            self._logger.msg.emit(
                f"Link established. Mapped data generated by <b>{source_item.name}</b> will be "
                f"imported in <b>{self.name}</b> when executing."
            )
        elif source_item.item_type() == "Tool":
            self._logger.msg.emit("Link established")
        elif source_item.item_type() == "Data Transformer":
            self._logger.msg.emit(
                "Link established. "
                f"Data transformed by <b>{source_item.name}</b> will be merged into <b>{self.name}</b> upon execution."
            )
        elif source_item.item_type() == "Merger":
            src_ds_names = ", ".join(x.name for x in source_item.predecessor_data_stores())
            self._logger.msg.emit(
                "Link established. "
                f"Data from <b>{src_ds_names}</b> will be merged into <b>{self.name}</b> upon execution."
            )
        else:
            super().notify_destination(source_item)

    def resources_for_direct_successors(self):
        """See base class."""
        if not self._url_validated:
            return []
        sa_url = convert_to_sqlalchemy_url(self._url, self.name)
        resources = scan_for_resources(self, sa_url)
        return resources

    def resources_for_direct_predecessors(self):
        """See base class."""
        return self.resources_for_direct_successors()

    def tear_down(self):
        """See base class"""
        if self._url_validated:
            sa_url = convert_to_sqlalchemy_url(self._url, self.name)
            self._toolbox.db_mngr.name_registry.unregister(sa_url, self.name)
        self._toolbox.db_mngr.database_clean_changed.disconnect(self._set_database_clean)
        self._database_validator.wait_for_finish()
        super().tear_down()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the :class:`MappingsTableProxy` model."""
from itertools import takewhile
from PySide6.QtCore import QSortFilterProxyModel


class MappingsTableProxy(QSortFilterProxyModel):
    def lessThan(self, source_left, source_right):
        if source_left.column() == 0:
            left_name = source_left.data()
            left_count = "".join(takewhile(lambda c: c.isdigit(), reversed(left_name)))
            if not left_count:
                return super().lessThan(source_left, source_right)
            right_name = source_right.data()
            right_count = "".join(takewhile(lambda c: c.isdigit(), reversed(right_name)))
            if not right_count:
                return super().lessThan(source_left, source_right)
            left_base = left_name[: len(left_name) - len(left_count)]
            right_base = right_name[: len(right_name) - len(right_count)]
            if left_base != right_base:
                return super().lessThan(source_left, source_right)
            return int(left_count[::-1]) < int(right_count[::-1])
        return source_left.row() < source_right.row()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities to scan for Data Connection's output resources."""
from pathlib import Path
from spine_engine.project_item.project_item_resource import file_resource, transient_file_resource, url_resource
from spine_engine.utils.serialization import path_in_dir
from spinedb_api.helpers import remove_credentials_from_url
from ..utils import convert_to_sqlalchemy_url


def scan_for_resources(provider, file_paths, urls, project_dir):
    """
    Creates file and URL resources based on DC's references and data.

    Args:
        provider (ProjectItem or ExecutableItem): resource provider item
        file_paths (list of str): file paths
        urls (list of dict): urls
        project_dir (str): absolute path to project directory

    Returns:
        list of ProjectItemResource: output resources
    """
    resources = []
    for fp in file_paths:
        try:
            if path_in_dir(fp, provider.data_dir):
                resource = file_resource(
                    provider.name, fp, label=f"<{provider.name}>/" + Path(fp).relative_to(provider.data_dir).as_posix()
                )
            elif path_in_dir(fp, project_dir):
                path = Path(fp)
                label = "<project>/" + Path(fp).relative_to(project_dir).as_posix()
                if path.exists():
                    resource = file_resource(provider.name, fp, label=label)
                else:
                    resource = transient_file_resource(provider.name, label)
            else:
                if Path(fp).exists():
                    resource = file_resource(provider.name, fp)
                else:
                    resource = transient_file_resource(provider.name, fp)
        except PermissionError:
            continue
        resources.append(resource)
    for url in urls:
        str_url = str(convert_to_sqlalchemy_url(url))
        schema = url.get("schema")
        resource = url_resource(
            provider.name, str_url, f"<{provider.name}>" + remove_credentials_from_url(str_url), schema=schema
        )
        resources.append(resource)
    return resources



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Contains utilities shared between project items. """
from contextlib import suppress
import os.path
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL, make_url
from spine_engine.utils.queue_logger import SuppressedMessage
import spinedb_api
from spinedb_api.filters.scenario_filter import scenario_name_from_dict
from spinedb_api.helpers import SUPPORTED_DIALECTS, UNSUPPORTED_DIALECTS, remove_credentials_from_url


class URLError(Exception):
    """Exception for errors in URL dicts."""


def database_label(provider_name):
    """Creates a standardized label for database resources.

    Args:
        provider_name (str): resource provider's name

    Returns:
        str: resource label
    """
    return "db_url@" + provider_name


def convert_to_sqlalchemy_url(url, item_name="", logger=None):
    """Returns a sqlalchemy url from url dict or None if not valid.

    Args:
        url (dict): URL to convert
        item_name (str): project item name for logging
        logger (LoggerInterface, optional): a logger

    Returns:
        URL: SqlAlchemy URL
    """
    selections = f"<b>{item_name}</b> selections" if item_name else "selections"
    if logger is None:
        logger = _NoLogger()
    if not url:
        logger.msg_error.emit(f"No URL specified for {selections}. Please specify one and try again.")
        return None
    try:
        sa_url = _convert_url(url)
        _validate_sa_url(sa_url, url["dialect"])
        return sa_url
    except URLError as error:
        logger.msg_error.emit(f"Unable to generate URL from {selections}: {error}")
        return None


def _convert_url(url):
    """Converts URL dict to SqlAlchemy URL.

    Args:
        url (dict): URL dictionary

    Returns:
        URL: SqlAlchemy URL
    """
    try:
        url = {key: value for key, value in url.items() if value}
        dialect = url.pop("dialect", None)
        with suppress(KeyError):
            del url["schema"]
        if not dialect:
            raise URLError("missing dialect")
        if dialect not in set(SUPPORTED_DIALECTS) | set(UNSUPPORTED_DIALECTS):
            raise URLError(f"invalid dialect '{dialect}'")
        if dialect == "sqlite":
            database = url.get("database", "")
            if database:
                url["database"] = os.path.abspath(database)
            return URL.create("sqlite", **url)  # pylint: disable=unexpected-keyword-arg
        db_api = spinedb_api.SUPPORTED_DIALECTS.get(dialect)
        if db_api is None:
            db_api = spinedb_api.helpers.UNSUPPORTED_DIALECTS[dialect]
        driver_name = f"{dialect}+{db_api}"
        return URL.create(driver_name, **url)  # pylint: disable=unexpected-keyword-arg
    except Exception as error:
        raise URLError(str(error)) from error


def _validate_sa_url(sa_url, dialect):
    """Validates SqlAlchemy URL.

    Args:
        sa_url (URL): SqlAlchemy URL to validate
        dialect (str): dialect

    Raises:
        URLError: raised if given URL is invalid
    """
    if not sa_url.database:
        raise URLError("database missing")
    if dialect != "sqlite":
        if sa_url.host is None:
            raise URLError("missing host")
        if sa_url.port is None:
            raise URLError("missing port")
        if sa_url.username is None:
            raise URLError("missing username")
        if sa_url.password is None:
            raise URLError("missing password")


def convert_url_to_safe_string(url):
    """Converts dict-style database URL to string without credentials.

    Args:
        url (dict): URL to convert

    Returns:
        str: URL as string
    """
    return remove_credentials_from_url(str(_convert_url(url)))


def check_database_url(sa_url):
    """Checks if URL points to a real database.

    Args:
        sa_url (URL): SQLAlchemy URL

    Returns:
        str: error message or None if URL works fine
    """
    try:
        engine = create_engine(sa_url)
        with engine.connect():
            pass
    except Exception as error:  # pylint: disable=broad-except
        return str(error)
    return None


class _NoLogger:
    def __getattr__(self, _name):
        return SuppressedMessage()


def split_url_credentials(url):
    """Pops username and password information from URL.

    Args:
        url (str): a URL

    Returns:
        tuple: URL without credentials and a tuple containing username, password pair
    """
    sa_url = make_url(url)
    connect_args = sa_url.translate_connect_args()
    username = connect_args.pop("username", None)
    password = connect_args.pop("password", None)
    new_sa_url = URL.create(sa_url.drivername, **connect_args)
    return str(new_sa_url), (username, password)


def unsplit_url_credentials(url, credentials):
    """Inserts username and password into a URL.

    Args:
        url (str): a URL
        credentials (tuple of str): username, password pair

    Returns:
        str: URL with credential information
    """
    sa_url = make_url(url)
    connect_args = sa_url.translate_connect_args()
    connect_args["username"], connect_args["password"] = credentials
    new_sa_url = URL.create(sa_url.drivername, **connect_args)
    return new_sa_url.render_as_string(hide_password=False)


def generate_filter_subdirectory_name(resources, filter_id_hash):
    """Generates an output directory name based on applied filters.

    Args:
        resources (Iterable of ProjectItemResource): item's forward resources
        filter_id_hash (str): hashed filter id string

    Returns:
        str: subdirectory name
    """
    subdirectory = filter_id_hash
    if subdirectory:
        scenario_name = _single_scenario_name_or_none(resources)
        if scenario_name is not None:
            subdirectory = scenario_name[:20] + "_" + subdirectory
    return subdirectory


def _single_scenario_name_or_none(resources):
    """Figures out if given resources have a single common scenario filter.

    Args:
        resources (Iterable of ProjectItemResource): resources to investigate

    Returns:
        str: scenario name or None
    """
    scenario_name = None
    for resource in resources:
        try:
            filter_stack = resource.metadata["filter_stack"]
        except KeyError:
            continue
        for filter_config in filter_stack:
            name = scenario_name_from_dict(filter_config)
            if name is None:
                continue
            if scenario_name is None:
                scenario_name = name
            elif name != scenario_name:
                return None
    return scenario_name


def escape_backward_slashes(string):
    """Escapes Windows directory separators.

    Args:
        string (str): string to escape

    Returns:
        str: escaped string
    """
    return string.replace("\\", "\\\\")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes and functions that can be shared among unit test modules."""
from contextlib import contextmanager
import os.path
from unittest.mock import MagicMock, patch
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QApplication, QMainWindow, QWidget
from spinetoolbox.ui_main import ToolboxUI


class MockQSettings:
    """Class for replacing an argument where e.g. class constructor requires an instance of QSettings.
    For example all ToolSpecification classes require a QSettings instance."""

    # noinspection PyMethodMayBeStatic, PyPep8Naming
    def value(self, key, defaultValue=""):
        """Returns the default value"""
        return defaultValue

    # noinspection PyPep8Naming
    def setValue(self, key, value):
        """Returns without modifying anything."""
        return


def create_mock_toolbox():
    mock_toolbox = MagicMock()
    mock_toolbox.msg = MagicMock()
    mock_toolbox.msg.attach_mock(MagicMock(), "emit")
    mock_toolbox.msg_warning = MagicMock()
    mock_toolbox.msg_warning.attach_mock(MagicMock(), "emit")
    mock_toolbox.undo_stack.push.side_effect = lambda cmd: cmd.redo()
    mock_toolbox.filtered_spec_factory_models.__getitem__.return_value = QStandardItemModel()
    return mock_toolbox


def create_mock_toolbox_with_mock_qsettings():
    mock_toolbox = create_mock_toolbox()
    mock_toolbox.qsettings().value.side_effect = return_default_value
    return mock_toolbox


def create_mock_project(project_dir):
    mock_project = MagicMock()
    mock_project.project_dir = project_dir
    mock_project.items_dir = os.path.join(project_dir, ".spinetoolbox", "items")
    mock_project.get_specification.side_effect = lambda x: None
    return mock_project


def mock_finish_project_item_construction(factory, project_item, mock_toolbox):
    icon = factory.make_icon(mock_toolbox)
    project_item.set_icon(icon)
    properties_widget = factory.make_properties_widget(mock_toolbox)
    project_item.set_properties_ui(properties_widget.ui)
    project_item.set_up()
    return properties_widget


def create_toolboxui():
    """Returns ToolboxUI for tests."""
    with (
        patch("spinetoolbox.ui_main.ToolboxUI.set_app_style") as mock_set_app_style,
        patch("spinetoolbox.plugin_manager.PluginManager.load_installed_plugins"),
    ):
        mock_set_app_style.return_value = True
        toolbox = ToolboxUI()
        toolbox._qsettings = MagicMock()
        toolbox._qsettings.value = MagicMock()
        toolbox._qsettings.value.side_effect = return_default_value
    return toolbox


def create_project(toolbox, project_dir):
    """Creates a project for the given ToolboxUI."""
    with (patch("spinetoolbox.ui_main.ToolboxUI.update_recent_projects"),):
        toolbox.create_project(project_dir)


def create_toolboxui_with_project(project_dir):
    """Returns ToolboxUI with a project instance where
    QSettings among others has been mocked."""
    with (
        patch("spinetoolbox.ui_main.ToolboxUI.set_app_style") as mock_set_app_style,
        patch("spinetoolbox.ui_main.ToolboxUI.save_project"),
        patch("spinetoolbox.plugin_manager.PluginManager.load_installed_plugins"),
    ):
        mock_set_app_style.return_value = True
        toolbox = ToolboxUI()
        toolbox._qsettings = MagicMock()
        toolbox._qsettings.value = MagicMock()
        toolbox._qsettings.value.side_effect = return_default_value
        toolbox.create_project(project_dir)
    return toolbox


def return_default_value(key, defaultValue=None):
    """Side effect function for QSettings.value() which returns defaultValue."""
    return defaultValue


def clean_up_toolbox(toolbox):
    """Cleans up toolbox and project."""
    if toolbox.project():
        toolbox.close_project(ask_confirmation=False)
        QApplication.processEvents()  # Makes sure Design view animations finish properly.
    toolbox.db_mngr.close_all_sessions()
    toolbox.db_mngr.clean_up()
    toolbox.db_mngr = None
    # Delete undo stack explicitly to prevent emitting certain signals well after ToolboxUI has been destroyed.
    toolbox.undo_stack.deleteLater()
    toolbox.deleteLater()


@contextmanager
def parent_widget():
    parent = QMainWindow()
    try:
        yield parent
    finally:
        parent.deleteLater()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains ImportSources widget and SourceDataTableMenu."""
from operator import itemgetter
import pickle
from PySide6.QtCore import (
    QItemSelection,
    QItemSelectionModel,
    QMimeData,
    QModelIndex,
    QObject,
    QPoint,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication
from spinedb_api.exception import InvalidMappingComponent
from spinedb_api.import_mapping.type_conversion import (
    IntegerSequenceDateTimeConvertSpec,
    StringConvertSpec,
    value_to_convert_spec,
)
from spinetoolbox.helpers import CharIconEngine
from ..commands import DeleteMapping, PasteMappings, PasteOptions, RestoreMappingsFromDict, SetColumnDefaultType
from ..mvcmodels.mappings_model import Role
from ..mvcmodels.source_data_table_model import SourceDataTableModel
from .custom_menus import SourceDataTableMenu, SourceListMenu
from .mime_types import MAPPING_LIST_MIME_TYPE, TABLE_OPTIONS_MIME_TYPE
from .options_widget import OptionsWidget
from .table_view_with_button_header import TYPE_TO_FONT_AWESOME_ICON

_EMPTY_DEFAULT_COLUMN_TYPE = "undefined"


class ImportSources(QObject):
    """
    Loads and controls the 'Sources' and 'Source data' part of the window.
    Loads the 'Mappings' part of the window as the user changes the source table.
    """

    preview_data_updated = Signal(int)

    def __init__(self, mappings_model, ui, undo_stack, parent):
        """
        Args:
            mappings_model (MappingsModel): mappings model
            ui (Any): import editor window's UI
            undo_stack (QUndoStack): undo stack
            parent (ImportEditorWindow): import editor window
        """
        super().__init__(parent)
        self._mappings_model = mappings_model
        self._undo_stack = undo_stack
        self._ui = ui
        self._stored_source_list_row = 0

        # state
        self._connector = None
        self._source_data_model = SourceDataTableModel(self)
        # create ui
        self._ui.source_data_table.setModel(self._source_data_model)
        self._ui.source_data_table.set_undo_stack(self._undo_stack, self._select_table_for_undo)
        self._ui_source_data_table_menu = SourceDataTableMenu(self._mappings_model, self._ui)
        self._ui_options_widget = OptionsWidget(self._undo_stack)
        self._ui_options_widget.setEnabled(False)
        self._ui.verticalLayout_source_options.addWidget(self._ui_options_widget)
        self._ui.source_data_table.verticalHeader().display_all = False
        self._fill_default_column_type_combo_box_items()
        # connect signals
        self._mappings_model.modelAboutToBeReset.connect(self._store_source_list_current_index)
        self._mappings_model.modelReset.connect(self._restore_source_list_current_index)
        self._mappings_model.dataChanged.connect(self._handle_mapping_data_changed)
        self._mappings_model.row_or_column_type_recommended.connect(self._source_data_model.set_type)
        self._mappings_model.multi_column_type_recommended.connect(self._source_data_model.set_all_column_types)
        self._ui_options_widget.options_changed.connect(lambda _: self._clear_source_data_model())
        self._ui_options_widget.about_to_undo.connect(self._select_table_for_undo)
        self._ui_options_widget.load_default_mapping_requested.connect(self._load_default_mapping)
        self._ui.source_list.customContextMenuRequested.connect(self.show_source_list_context_menu)
        self._ui.source_list.selectionModel().currentChanged.connect(self._change_selected_table)
        self._ui.mapping_list.selectionModel().currentChanged.connect(self._change_selected_mapping)
        self._ui.source_data_table.customContextMenuRequested.connect(self._ui_source_data_table_menu.request_menu)
        self._ui.default_column_type_combo_box.currentTextChanged.connect(self._set_default_column_type)

        self._source_data_model.mapping_data_changed.connect(self._update_display_row_types)
        self._source_data_model.column_types_updated.connect(self._new_column_types)
        self._source_data_model.row_types_updated.connect(self._new_row_types)
        self._source_data_model.polish_mapping_requested.connect(self._polish_mappings_in_list)

    @Slot(QModelIndex, QModelIndex, list)
    def _handle_mapping_data_changed(self, top_left, bottom_right, roles):
        self._update_source_table_colors(top_left, bottom_right, roles)
        table_index = self._ui.source_list.selectionModel().currentIndex()
        if not table_index.isValid() or table_index.row() == 0:
            return
        header = self._source_data_model.header
        for list_row in range(self._mappings_model.rowCount(table_index)):
            list_index = self._mappings_model.index(list_row, 0, table_index)
            msg = self._mappings_model.check_validity_of_columns(list_index, header)
            if msg:
                self.parent().show_error(msg)
                return

    def set_connector(self, connector, mapping):
        """Sets connector.

        Args:
            connector (ConnectionManager): connector
            mapping (dict)
        """
        self._ui.source_list.selectionModel().clearCurrentIndex()
        self._connector = connector
        self._connector.connection_ready.connect(self.request_new_tables_from_connector)
        self._connector.data_ready.connect(self._update_source_data)
        self._connector.tables_ready.connect(self.update_tables)
        self._connector.default_mapping_ready.connect(self._set_default_mapping)
        self._ui_options_widget.set_connector(self._connector)
        self._source_data_model.more_data_needed.connect(self.fetch_more_data, Qt.UniqueConnection)
        self.restore_connectors(mapping)

    def _fill_default_column_type_combo_box_items(self):
        """Adds items to default column type combo box."""
        combo_box = self._ui.default_column_type_combo_box
        combo_box.addItem(_EMPTY_DEFAULT_COLUMN_TYPE)
        for data_type_label, icon_character in sorted(TYPE_TO_FONT_AWESOME_ICON.items(), key=itemgetter(0)):
            if data_type_label == IntegerSequenceDateTimeConvertSpec.DISPLAY_NAME:
                # Skipping the more complicated conversion for now.
                continue
            engine = CharIconEngine(icon_character, 0)
            icon = QIcon(engine.pixmap())
            self._ui.default_column_type_combo_box.addItem(icon, data_type_label)
        self._ui.default_column_type_combo_box.setCurrentText(_EMPTY_DEFAULT_COLUMN_TYPE)

    @Slot()
    def _polish_mappings_in_list(self):
        """Polishes mappings in mapping list."""
        table_index = self._ui.source_list.selectionModel().currentIndex()
        if not table_index.isValid() or table_index.row() == 0:
            return
        header = self._source_data_model.header
        for list_row in range(self._mappings_model.rowCount(table_index)):
            list_index = self._mappings_model.index(list_row, 0, table_index)
            try:
                self._mappings_model.polish_mapping(list_index, header)
            except InvalidMappingComponent as error:
                self.parent().show_error(str(error))

    @Slot(str)
    def _select_table_for_undo(self, table_name):
        """Selects source table to load correct values to different widgets.

        Args:
            table_name (str): table name
        """
        if table_name == self._connector.current_table:
            return
        for row in range(self._mappings_model.rowCount()):
            table_index = self._mappings_model.index(row, 0)
            if table_index.data() == table_name:
                self._ui.source_list.selectionModel().setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)
                break

    @Slot()
    def _load_default_mapping(self):
        self._connector.request_default_mapping()

    @Slot(dict)
    def _set_default_mapping(self, mapping):
        self._undo_stack.push(RestoreMappingsFromDict(self, self._mappings_model, mapping))

    @Slot()
    def request_new_tables_from_connector(self):
        """
        Requests new tables data from connector
        """
        self._connector.request_tables()

    @Slot(QModelIndex, QModelIndex)
    def _change_selected_table(self, current, _previous):
        """
        Sets selected table and requests data from connector

        Args:
            current (QModelIndex): current index
            _previous (QModelIndex): previous index
        """
        if self._connector is None:
            self._ui_options_widget.setEnabled(False)
            return
        if not current.isValid():
            table_name = ""
            self._ui_options_widget.setEnabled(False)
            self._ui.default_column_type_combo_box.setEnabled(False)
        else:
            table_item = self._mappings_model.data(current, Role.ITEM)
            table_name = table_item.name if table_item.real else ""
            self._ui_options_widget.setEnabled(bool(table_name))
            self._ui.default_column_type_combo_box.setEnabled(bool(table_name))
            if table_name:
                self._reset_default_column_type(table_name)
        self._connector.set_table(table_name)
        self._clear_source_data_model()

    @Slot(QModelIndex, QModelIndex)
    def _change_selected_mapping(self, current, previous):
        """Updates source table according to newly selected mapping from mapping list.

        Args:
            current (QModelIndex): current mapping
            previous (QModelIndex): previous mapping
        """
        self._source_data_model.set_mapping_list_index(current)

    @Slot(str)
    def _set_default_column_type(self, selected_label):
        """Pushes a set default column type command to undo stack.

        Args:
            selected_label (str): new default column data type
        """
        table_name = self._connector.current_table
        current_type = self._connector.table_default_column_type.get(table_name, _EMPTY_DEFAULT_COLUMN_TYPE)
        self._undo_stack.push(SetColumnDefaultType(self, table_name, selected_label, current_type))

    def change_default_column_type(self, table_name, column_type):
        """Sets default column type.

        Args:
            table_name (str): table name
            column_type (str): new default column type
        """
        self._select_table_for_undo(table_name)
        if column_type != _EMPTY_DEFAULT_COLUMN_TYPE:
            self._connector.update_table_default_column_type({table_name: value_to_convert_spec(column_type)})
        else:
            self._connector.clear_table_default_column_type(table_name)
        if column_type != self._ui.default_column_type_combo_box.currentText():
            self._reset_default_column_type(table_name)

    def _reset_default_column_type(self, table_name):
        """Updates the default column type combo box.

        Args:
            table_name (str): current_table_name
        """
        converter = self._connector.table_default_column_type.get(table_name)
        column_type = converter.DISPLAY_NAME if converter is not None else _EMPTY_DEFAULT_COLUMN_TYPE
        self._ui.default_column_type_combo_box.currentTextChanged.disconnect(self._set_default_column_type)
        self._ui.default_column_type_combo_box.setCurrentText(column_type)
        self._ui.default_column_type_combo_box.currentTextChanged.connect(self._set_default_column_type)

    def _update_source_table_colors(self, top_left, bottom_right, roles):
        """Notifies source table model that colors have changed.

        Args:
            top_left (QModelIndex): changed data's top left index in mappings model
            bottom_right (QModelIndex): changed data's bottom right index in mappings model
        """
        mapping_list_index = self._ui.mapping_list.selectionModel().currentIndex()
        if not mapping_list_index.isValid() or (
            (
                QItemSelection(top_left, bottom_right).contains(mapping_list_index)
                and Role.FLATTENED_MAPPINGS not in roles
            )
            and top_left.parent() != mapping_list_index
        ):
            return
        self._source_data_model.handle_mapping_data_changed()

    @Slot(int, int)
    def fetch_more_data(self, max_rows, start):
        self._connector.request_data(max_rows=max_rows, start=start)

    def _select_table_row(self, row):
        selection_model = self._ui.source_list.selectionModel()
        index = self._mappings_model.index(row, 0)
        selection_model.setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)

    @Slot(list)
    def update_tables(self, tables):
        """Updates list of tables.

        Args:
            tables (list of str): updated source tables
        """
        if self.parent().is_file_less():
            self._mappings_model.add_empty_row()
            return
        selection_model = self._ui.source_list.selectionModel()
        current_row = selection_model.currentIndex().row()
        self._mappings_model.cross_check_source_table_names(set(tables))
        self._mappings_model.remove_tables_not_in_source_and_specification()
        table_names = set(self._mappings_model.real_table_names())
        for t_name in tables:
            if t_name not in table_names:
                self._mappings_model.append_new_table_with_mapping(t_name, None)
        # reselect current table if existing otherwise select first table
        selection_model.setCurrentIndex(QModelIndex(), QItemSelectionModel.ClearAndSelect)
        if current_row >= 0:
            self._select_table_row(min(current_row, self._mappings_model.rowCount() - 1))
        else:
            self._select_table_row(1 if self._mappings_model.rowCount() > 1 else 0)

    @Slot(list, list)
    def _update_source_data(self, data, header):
        if not data and not header:
            return
        if data:
            try:
                data = _sanitize_data(data, header)
            except RuntimeError as error:
                self.parent().show_error(str(error))
                return
            self._source_data_model.append_rows(data)
            if not header:
                header = list(range(1, len(data[0]) + 1))
        # Set header data before resetting model because the header needs to be there for some slots...
        self._source_data_model.set_horizontal_header_labels(header)
        table_name = self._connector.current_table
        types = self._connector.table_types.get(table_name, {})
        row_types = self._connector.table_row_types.get(table_name, {})
        default_column_type = None
        for col in range(len(header)):
            col_type = types.get(col)
            if col_type is None:
                if default_column_type is None:
                    default_column_type = self._connector.table_default_column_type.get(table_name)
                    if default_column_type is None:
                        default_column_type = StringConvertSpec()
                col_type = default_column_type.DISPLAY_NAME
            self._source_data_model.set_type(
                col, value_to_convert_spec(col_type), orientation=Qt.Orientation.Horizontal
            )
        for row, row_type in row_types.items():
            self._source_data_model.set_type(row, value_to_convert_spec(row_type), orientation=Qt.Orientation.Vertical)
        self.preview_data_updated.emit(self._source_data_model.columnCount())

    def _clear_source_data_model(self):
        self._source_data_model.clear(infinite=self.parent().is_file_less())
        self.preview_data_updated.emit(0)

    def restore_connectors(self, mappings_dict):
        """Restores connectors.

        Args:
            mappings_dict (dict): serialized data
        """
        table_types = {
            tn: {int(col): value_to_convert_spec(spec) for col, spec in cols.items()}
            for tn, cols in mappings_dict.get("table_types", {}).items()
        }
        table_default_column_type = {
            tn: value_to_convert_spec(spec) for tn, spec in mappings_dict.get("table_default_column_type", {}).items()
        }
        table_row_types = {
            tn: {int(row): value_to_convert_spec(spec) for row, spec in rows.items()}
            for tn, rows in mappings_dict.get("table_row_types", {}).items()
        }
        self._connector.set_table_options(mappings_dict.get("table_options", {}))
        self._connector.set_table_types(table_types)
        self._connector.update_table_default_column_type(table_default_column_type)
        self._connector.set_table_row_types(table_row_types)

    @Slot()
    def _store_source_list_current_index(self):
        """Stores source table list's current index."""
        self._stored_source_list_row = self._ui.source_list.selectionModel().currentIndex().row()

    @Slot()
    def _restore_source_list_current_index(self):
        """Restores source table list's current index."""
        row = min(self._stored_source_list_row, self._mappings_model.rowCount() - 1)
        current = self._mappings_model.index(row, 0)
        self._ui.source_list.selectionModel().setCurrentIndex(current, QItemSelectionModel.ClearAndSelect)

    def store_connectors(self):
        """Returns a dictionary with type of connector and connector options for tables.

        Returns:
            dict: stored connectors
        """
        table_names = set(self._mappings_model.real_table_names())
        table_types = {
            tn: {col: spec.to_json_value() for col, spec in cols.items()}
            for tn, cols in self._connector.table_types.items()
            if cols
            if tn in table_names
        }
        table_default_column_type = {
            tn: column_type.to_json_value()
            for tn, column_type in self._connector.table_default_column_type.items()
            if tn in table_names
        }
        table_row_types = {
            tn: {col: spec.to_json_value() for col, spec in cols.items()}
            for tn, cols in self._connector.table_row_types.items()
            if cols and tn in table_names
        }
        table_options = {t: o for t, o in self._connector.table_options.items() if t in table_names}
        return {
            "table_options": table_options,
            "table_types": table_types,
            "table_default_column_type": table_default_column_type,
            "table_row_types": table_row_types,
            "source_type": self._connector.source_type,
        }

    @Slot()
    def close_connection(self):
        """Close connector connection."""
        if self._connector is not None:
            self._connector.close_connection()

    @Slot()
    def _new_column_types(self):
        new_types = self._source_data_model.get_types(orientation=Qt.Orientation.Horizontal)
        self._connector.set_table_types({self._connector.current_table: new_types})

    @Slot()
    def _new_row_types(self):
        new_types = self._source_data_model.get_types(orientation=Qt.Orientation.Vertical)
        self._connector.set_table_row_types({self._connector.current_table: new_types})

    @Slot()
    def _update_display_row_types(self):
        """Updates displayed row types."""
        list_index = self._ui.mapping_list.selectionModel().currentIndex()
        if not list_index.isValid():
            return
        flattened_mapping = list_index.data(Role.FLATTENED_MAPPINGS)
        last_pivot_row = flattened_mapping.root_mapping.last_pivot_row()
        if last_pivot_row == -1:
            pivoted_rows = []
        else:
            pivoted_rows = list(range(last_pivot_row + 1))
        self._ui.source_data_table.verticalHeader().sections_with_buttons = pivoted_rows

    @Slot(QPoint)
    def show_source_list_context_menu(self, pos):
        """
        Shows context menu for source tables.

        Args:
            pos (QPoint): Mouse position
        """
        global_pos = self._ui.source_list.mapToGlobal(pos)
        index = self._ui.source_list.indexAt(pos)
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        source_list_menu = SourceListMenu(
            self._ui.source_list,
            global_pos,
            mime_data.hasFormat(TABLE_OPTIONS_MIME_TYPE),
            mime_data.hasFormat(MAPPING_LIST_MIME_TYPE),
        )
        option = source_list_menu.get_action()
        source_list_menu.deleteLater()
        if option == "Copy mappings":
            self._copy_mapping_list_to_clipboard(index)
            return
        if option == "Copy options":
            self._copy_table_options_to_clipboard(index)
            return
        if option == "Copy options and mappings":
            self._copy_mappings_and_options_to_clipboard(index)
            return
        if option == "Paste mappings":
            self._paste_mapping_list_from_clipboard()
            return
        if option == "Paste options":
            self._paste_table_options_from_clipboard()
            return
        if option == "Paste options and mappings":
            self._undo_stack.beginMacro("paste options and mappings")
            self._paste_mapping_list_from_clipboard()
            self._paste_table_options_from_clipboard()
            self._undo_stack.endMacro()

    def _copy_mapping_list_to_clipboard(self, table_index):
        """
        Copies the mappings of given source table to system clipboard.

        Args:
            table_index (QModelIndex): table index
        """
        indexes = [
            self._mappings_model.index(row, 0, table_index) for row in range(self._mappings_model.rowCount(table_index))
        ]
        mime_data = self._mappings_model.mimeData(indexes)
        clipboard = QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def _paste_mapping_list_from_clipboard(self):
        """Pastes mapping list from system clipboard to selected source tables."""
        selection_model = self._ui.source_list.selectionModel()
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        mapping_list_data = pickle.loads(mime_data.data(MAPPING_LIST_MIME_TYPE))
        self._undo_stack.beginMacro("paste mappings")
        for table_index in selection_model.selectedIndexes():
            table_row = table_index.row()
            if table_row == 0:
                continue
            for list_row in reversed(range(self._mappings_model.rowCount(table_index))):
                self._undo_stack.push(DeleteMapping(table_row, self._mappings_model, list_row))
            target_row = self._mappings_model.rowCount(table_index)
            self._undo_stack.push(PasteMappings(table_row, target_row, self._mappings_model, mapping_list_data))
        self._undo_stack.endMacro()

    def _copy_table_options_to_clipboard(self, table_index):
        """
        Serializes mapping options to a dict.

        Args:
            table_index (QModelIndex): table index
        """
        mime_data = QMimeData()
        mime_data.setData(TABLE_OPTIONS_MIME_TYPE, self._pickle_table_options(table_index))
        clipboard = QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def _pickle_table_options(self, table_index):
        """Pickles options from given table.

        Args:
            table_index (QModelIndex): table index

        Returns:
            bytes: pickles options
        """
        options = self._connector.table_options
        col_types = self._connector.table_types
        row_types = self._connector.table_row_types
        table_name = table_index.data(Role.ITEM).name
        options_dict = {}
        options_dict["options"] = options.get(table_name, {})
        options_dict["col_types"] = col_types.get(table_name, {})
        options_dict["row_types"] = row_types.get(table_name, {})
        return pickle.dumps(options_dict)

    def _paste_table_options_from_clipboard(self):
        selection_model = self._ui.source_list.selectionModel()
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        pickled_options = mime_data.data(TABLE_OPTIONS_MIME_TYPE)
        self._undo_stack.beginMacro("paste table options")
        for table_index in selection_model.selectedIndexes():
            table_row = table_index.row()
            if table_row == 0:
                continue
            pickled_previous_options = self._pickle_table_options(table_index)
            table_name = table_index.data(Role.ITEM).name
            self._undo_stack.push(PasteOptions(self, table_name, pickled_options, pickled_previous_options))
        self._undo_stack.endMacro()

    def paste_options(self, table, options):
        """
        Pastes all mapping options to given table.

        Args:
            table (str): source table name
            options (dict): options
        """
        self._connector.set_table_options({table: options.get("options", {})})
        self._connector.set_table_types({table: options.get("col_types", {})})
        self._connector.set_table_row_types({table: options.get("row_types", {})})

    def _copy_mappings_and_options_to_clipboard(self, table_index):
        """Copies table's mapping list and options to system clipboard.

        Args:
            table_index (QModelIndex): table index
        """
        indexes = [
            self._mappings_model.index(row, 0, table_index) for row in range(self._mappings_model.rowCount(table_index))
        ]
        mime_data = self._mappings_model.mimeData(indexes)
        mime_data.setData(TABLE_OPTIONS_MIME_TYPE, self._pickle_table_options(table_index))
        clipboard = QApplication.clipboard()
        clipboard.setMimeData(mime_data)


def _sanitize_data(data, header):
    """Fills empty data cells with None."""
    expected_columns = len(header) if header else max((len(x) for x in data))
    sanitized_data = []
    for row in data:
        length_diff = expected_columns - len(row)
        if length_diff > 0:
            row = row + length_diff * [None]
        elif length_diff < 0:
            raise RuntimeError("Data row has more items than the header row.")
        sanitized_data.append(row)
    return sanitized_data



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Undo/redo commands for the View project item."""
from copy import deepcopy
from spine_items.commands import SpineToolboxCommand


class PinOrUnpinDBValuesCommand(SpineToolboxCommand):
    """Command to pin or unpin DB values."""

    def __init__(self, view_name, new_values, old_values, project):
        """
        Args:
            view_name (str): View's name
            new_values (dict): mapping name to list of value identifiers
            old_values (dict): mapping name to list of value identifiers
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._view_name = view_name
        self._new_values = deepcopy(new_values)
        self._old_values = deepcopy(old_values)
        self._project = project
        texts = ["pin" if values else "unpin" + " " + name for name, values in new_values.items()]
        self.setText(f"{', '.join(texts)} in {view_name}")

    def redo(self):
        view = self._project.get_item(self._view_name)
        view.do_pin_db_values(deepcopy(self._new_values))

    def undo(self):
        view = self._project.get_item(self._view_name)
        view.do_pin_db_values(deepcopy(self._old_values))



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for Import editor's :class:`MappingsModel`."""
import unittest
from PySide6.QtCore import QModelIndex, QObject, Qt
from PySide6.QtGui import QUndoStack
from PySide6.QtWidgets import QApplication
from spine_items.importer.commands import DeleteMapping
from spine_items.importer.mvcmodels.mappings_model import MappingsModel
from spine_items.importer.mvcmodels.mappings_model_roles import Role
from spinedb_api import import_mapping_from_dict
from spinetoolbox.helpers import signal_waiter
from tests.importer.helpers import append_source_table_with_mappings


class TestMappingsModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._model_parent = QObject()
        self._undo_stack = QUndoStack(self._model_parent)

    def tearDown(self):
        self._model_parent.deleteLater()
        QApplication.processEvents()

    def test_empty_model_has_Select_All_item(self):
        model = MappingsModel(self._undo_stack, self._model_parent)
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.columnCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "Select all")

    def test_set_time_series_repeat_flag(self):
        model = MappingsModel(self._undo_stack, self._model_parent)
        model.add_empty_row()
        table_index = model.index(1, 0)
        with signal_waiter(model.dataChanged) as waiter:
            self.assertTrue(model.setData(table_index, "table"))
            waiter.wait()
        self.assertTrue(model.insertRow(0, table_index))
        list_index = model.index(0, 0, table_index)
        model.set_root_mapping(
            table_index.row(), list_index.row(), import_mapping_from_dict({"map_type": "ObjectClass"})
        )
        model.set_parameter_type(table_index.row(), list_index.row(), "Value")
        model.set_value_type(table_index.row(), list_index.row(), "Time series")
        flattened_mappings = model.data(list_index, Role.FLATTENED_MAPPINGS)
        self.assertNotIn("repeat", flattened_mappings.value_mapping().options)
        with signal_waiter(model.dataChanged) as waiter:
            model.set_time_series_repeat_flag(table_index.row(), list_index.row(), True)
            waiter.wait()
        self.assertTrue(flattened_mappings.value_mapping().options["repeat"])

    def test_change_mappings_type(self):
        model = MappingsModel(self._undo_stack, self._model_parent)
        model.restore(
            {
                "table_mappings": {
                    "Sheet1": [
                        {
                            "Mapping 1": {
                                "mapping": [
                                    {"map_type": "EntityClass", "position": "hidden", "value": "Object"},
                                    {"map_type": "Entity", "position": 0},
                                    {"map_type": "EntityMetadata", "position": "hidden"},
                                    {"map_type": "ParameterDefinition", "position": "hidden", "value": "size"},
                                    {"map_type": "Alternative", "position": "hidden", "value": "Base"},
                                    {"map_type": "ParameterValueMetadata", "position": "hidden"},
                                    {"map_type": "ParameterValue", "position": 1},
                                ]
                            }
                        }
                    ]
                },
                "selected_tables": ["Sheet1"],
                "table_options": {"Sheet1": {}},
                "table_types": {"Sheet1": {"0": "string", "1": "float"}},
                "table_default_column_type": {},
                "table_row_types": {},
                "source_type": "ExcelConnector",
            }
        )
        self.assertEqual(model.index(0, 0).data(), "Select all")
        table_index = model.index(1, 0)
        self.assertEqual(table_index.data(), "Sheet1")
        list_index = model.index(0, 0, table_index)
        self.assertEqual(list_index.data(), "Mapping 1")
        expected = [
            ["Entity class names", "Constant", "Object", ""],
            ["Entity names", "Column", 1, ""],
            ["Entity metadata", "None", None, ""],
            ["Parameter names", "Constant", "size", ""],
            ["Alternative names", "Constant", "Base", ""],
            ["Parameter value metadata", "None", None, ""],
            ["Parameter values", "Column", 2, ""],
        ]
        rows = model.rowCount(list_index)
        self.assertEqual(rows, len(expected))
        for row in range(model.rowCount(list_index)):
            expected_row = expected[row]
            columns = model.columnCount(list_index)
            self.assertEqual(columns, len(expected_row))
            for column in range(columns):
                with self.subTest(row=row, column=column):
                    index = model.index(row, column, list_index)
                    self.assertEqual(index.data(), expected_row[column])
        model.set_mappings_type(1, 0, "Entity group")
        expected = [
            ["Entity class names", "None", None, ""],
            ["Group names", "None", None, ""],
            ["Member names", "None", None, ""],
        ]
        rows = model.rowCount(list_index)
        self.assertEqual(rows, len(expected))
        for row in range(model.rowCount(list_index)):
            expected_row = expected[row]
            columns = model.columnCount(list_index)
            self.assertEqual(columns, len(expected_row))
            for column in range(columns):
                with self.subTest(row=row, column=column):
                    index = model.index(row, column, list_index)
                    self.assertEqual(index.data(), expected_row[column])


class TestTableList(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._model_parent = QObject()
        self._undo_stack = QUndoStack(self._model_parent)
        self._model = MappingsModel(self._undo_stack, self._model_parent)

    def tearDown(self):
        self._model_parent.deleteLater()
        QApplication.processEvents()

    def test_set_tables_editable(self):
        self._model.add_empty_row()
        table_index = self._model.index(1, 0)
        with signal_waiter(self._model.dataChanged) as waiter:
            self.assertTrue(self._model.setData(table_index, "table"))
            waiter.wait()
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(int((self._model.flags(self._model.index(0, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)
        self.assertNotEqual(int((self._model.flags(self._model.index(1, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)
        self.assertNotEqual(int((self._model.flags(self._model.index(2, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)
        self._model.set_tables_editable(False)
        self.assertEqual(int((self._model.flags(self._model.index(0, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)
        self.assertEqual(int((self._model.flags(self._model.index(1, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)
        self.assertNotEqual(int((self._model.flags(self._model.index(2, 0)) & Qt.ItemFlag.ItemIsEditable).value), 0)

    def test_set_source_table_items_into_specification(self):
        self._model.append_new_table_with_mapping("my source table", None)
        self.assertEqual(self._model.rowCount(), 2)
        for row in range(1, self._model.rowCount()):
            item = self._model.index(row, 0).data(Role.ITEM)
            self.assertFalse(item.in_specification)
        self._model.set_source_table_items_into_specification()
        for row in range(1, self._model.rowCount()):
            item = self._model.index(row, 0).data(Role.ITEM)
            self.assertTrue(item.in_specification)

    def test_remove_tables_not_in_source_and_specification(self):
        self._model.append_new_table_with_mapping("table that gets removed", None)
        self._model.append_new_table_with_mapping("table that is only in source", None)
        root_mapping = import_mapping_from_dict({"map_type": "ObjectClass", "name": None, "object": None})
        self._model.append_new_table_with_mapping("table that is in source and specification", root_mapping)
        self.assertEqual(self._model.rowCount(), 4)
        self._model.cross_check_source_table_names(
            {"table that is only in source", "table that is in source and specification"}
        )
        self._model.remove_tables_not_in_source_and_specification()
        self.assertEqual(self._model.rowCount(), 3)
        expected_names = ["Select all", "table that is only in source", "table that is in source and specification"]
        for row, expected_name in zip(range(self._model.rowCount()), expected_names):
            item = self._model.index(row, 0).data(Role.ITEM)
            self.assertEqual(item.name, expected_name)

    def test_cross_check_source_table_names(self):
        self._model.append_new_table_with_mapping("initially in source", None)
        self._model.append_new_table_with_mapping("initially not in source", None)
        not_in_source_item = self._model.index(self._model.rowCount() - 1, 0).data(Role.ITEM)
        self.assertEqual(not_in_source_item.name, "initially not in source")
        not_in_source_item.in_source = False
        self._model.cross_check_source_table_names({"initially not in source"})
        self.assertEqual(self._model.rowCount(), 3)
        expected_in_source = {"initially in source": False, "initially not in source": True}
        expected_real = {"initially in source": False, "initially not in source": True}
        for row in range(1, self._model.rowCount()):
            item = self._model.index(row, 0).data(Role.ITEM)
            self.assertEqual(item.in_source, expected_in_source[item.name])
            self.assertEqual(item.real, expected_real[item.name])

    def test_empty_model_has_select_all_source_table_item(self):
        self.assertEqual(self._model.rowCount(), 1)
        index = self._model.index(0, 0)
        self.assertEqual(index.data(), "Select all")
        self.assertIsNone(index.data(Qt.ItemDataRole.ForegroundRole))
        self.assertEqual(index.data(Qt.ItemDataRole.CheckStateRole), Qt.CheckState.Checked)
        self.assertIsNone(index.data(Qt.ItemDataRole.FontRole))
        self.assertIsNone(index.data(Qt.ItemDataRole.ToolTipRole))
        flags = self._model.flags(index)
        self.assertEqual(flags, Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)

    def test_empty_row(self):
        self._model.add_empty_row()
        self.assertEqual(self._model.rowCount(), 2)
        index = self._model.index(1, 0)
        self._assert_empty_row(index)

    def test_turn_empty_row_into_non_real_table(self):
        self._model.add_empty_row()
        empty_row_index = self._model.index(1, 0)
        with signal_waiter(self._model.dataChanged, timeout=1.0) as waiter:
            self._model.setData(empty_row_index, "my shiny table")
            waiter.wait()
            self.assertEqual(len(waiter.args), 3)
            self.assertTrue(self._model.is_table_index(waiter.args[0]))
            self.assertEqual(waiter.args[0].row(), 1)
            self.assertEqual(waiter.args[0].column(), 0)
            self.assertTrue(self._model.is_table_index(waiter.args[1]))
            self.assertEqual(waiter.args[1].row(), 1)
            self.assertEqual(waiter.args[1].column(), 0)
        self.assertEqual(self._model.rowCount(), 3)
        index = self._model.index(1, 0)
        self.assertEqual(index.data(), "my shiny table (new)")
        self.assertIsNone(index.data(Qt.ItemDataRole.ForegroundRole))
        self.assertEqual(index.data(Qt.ItemDataRole.CheckStateRole), Qt.CheckState.Checked)
        self.assertIsNone(index.data(Qt.ItemDataRole.FontRole))
        self.assertEqual(
            index.data(Qt.ItemDataRole.ToolTipRole),
            "<qt>Table's mappings haven't been saved with the specification yet.</qt>",
        )
        flags = self._model.flags(index)
        self.assertEqual(
            flags,
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsDragEnabled
            | Qt.ItemFlag.ItemIsDropEnabled,
        )
        index = self._model.index(2, 0)
        self._assert_empty_row(index)

    def test_when_empty_row_turns_into_non_real_table_its_check_status_equals_select_all(self):
        self._model.add_empty_row()
        index = self._model.index(0, 0)
        self.assertTrue(self._model.setData(index, Qt.CheckState.Unchecked.value, Qt.ItemDataRole.CheckStateRole))
        empty_row_index = self._model.index(1, 0)
        with signal_waiter(self._model.dataChanged, timeout=1.0) as waiter:
            self._model.setData(empty_row_index, "my shiny table")
            waiter.wait()
        self.assertEqual(self._model.rowCount(), 3)
        index = self._model.index(1, 0)
        self.assertEqual(index.data(), "my shiny table (new)")
        self.assertEqual(index.data(Qt.ItemDataRole.CheckStateRole), Qt.CheckState.Unchecked)
        index = self._model.index(2, 0)
        self._assert_empty_row(index)

    def _assert_empty_row(self, index):
        self.assertEqual(index.data(), "<rename this to add table>")
        self.assertEqual(index.data(Qt.ItemDataRole.CheckStateRole), Qt.CheckState.Unchecked)
        self.assertIsNone(index.data(Qt.ItemDataRole.ForegroundRole))
        self.assertTrue(index.data(Qt.ItemDataRole.FontRole).italic())
        flags = self._model.flags(index)
        self.assertEqual(flags, Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable)

    def test_undo_create_new_table_restores_empty_row(self):
        self._model.add_empty_row()
        empty_row_index = self._model.index(1, 0)
        with signal_waiter(self._model.dataChanged, timeout=1.0) as waiter:
            self._model.setData(empty_row_index, "my temporary table")
            waiter.wait()
        self.assertEqual(self._model.rowCount(), 3)
        with signal_waiter(self._model.dataChanged, timeout=1.0) as waiter:
            self._undo_stack.undo()
            waiter.wait()
        self.assertEqual(self._model.rowCount(), 2)
        index = self._model.index(1, 0)
        self._assert_empty_row(index)

    def test_move_source_table_rows(self):
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        append_source_table_with_mappings(self, self._model, "two", "second", [])
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "one")
        self.assertEqual(self._model.index(2, 0).data(), "two")
        self.assertTrue(self._model.moveRows(QModelIndex(), 1, 1, QModelIndex(), 3))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "two")
        self.assertEqual(self._model.index(2, 0).data(), "one")
        self.assertTrue(self._model.moveRows(QModelIndex(), 2, 1, QModelIndex(), 1))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0, QModelIndex()).data(), "one")
        self.assertEqual(self._model.index(2, 0, QModelIndex()).data(), "two")

    def test_internal_move_source_table_row_by_mime_data_drop(self):
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        append_source_table_with_mappings(self, self._model, "two", "second", [])
        mime_data = self._model.mimeData([self._model.index(1, 0)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 3, 0, QModelIndex()))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "two")
        self.assertEqual(
            self._model.index(
                2,
                0,
            ).data(),
            "one",
        )
        mime_data = self._model.mimeData([self._model.index(2, 0)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 1, 0, QModelIndex()))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "one")
        self.assertEqual(self._model.index(2, 0).data(), "two")

    def test_cannot_drop_source_tables_above_select_all_row(self):
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        mime_data = self._model.mimeData([self._model.index(1, 0)])
        self.assertFalse(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 0, 0, QModelIndex()))

    def test_cannot_drop_source_tables_below_empty_row(self):
        self._model.add_empty_row()
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        self.assertTrue(self._model.has_empty_source_table_row())
        self.assertEqual(self._model.rowCount(), 3)
        mime_data = self._model.mimeData([self._model.index(1, 0)])
        self.assertFalse(
            self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, self._model.rowCount(), 0, QModelIndex())
        )

    def test_undo_moving_source_table_row_down_by_mime_data_drop(self):
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        append_source_table_with_mappings(self, self._model, "two", "second", [])
        mime_data = self._model.mimeData([self._model.index(1, 0)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 3, 0, QModelIndex()))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "two")
        self.assertEqual(
            self._model.index(
                2,
                0,
            ).data(),
            "one",
        )
        self._undo_stack.undo()
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "one")
        self.assertEqual(
            self._model.index(
                2,
                0,
            ).data(),
            "two",
        )

    def test_undo_moving_source_table_row_up_by_mime_data_drop(self):
        append_source_table_with_mappings(self, self._model, "one", "first", [])
        append_source_table_with_mappings(self, self._model, "two", "second", [])
        mime_data = self._model.mimeData([self._model.index(2, 0)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 1, 0, QModelIndex()))
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "two")
        self.assertEqual(
            self._model.index(
                2,
                0,
            ).data(),
            "one",
        )
        self._undo_stack.undo()
        self.assertEqual(self._model.rowCount(), 3)
        self.assertEqual(self._model.index(0, 0, QModelIndex()).data(), "Select all")
        self.assertEqual(self._model.index(1, 0).data(), "one")
        self.assertEqual(
            self._model.index(
                2,
                0,
            ).data(),
            "two",
        )


class TestMappingList(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._model_parent = QObject()
        self._undo_stack = QUndoStack(self._model_parent)
        self._model = MappingsModel(self._undo_stack, self._model_parent)

    def _append_source_table_with_mappings(self, table_name, first_mapping_name, additional_mapping_names):
        return append_source_table_with_mappings(
            self, self._model, table_name, first_mapping_name, additional_mapping_names
        )

    def test_move_mapping_row(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second"])
        self.assertTrue(self._model.moveRows(table_index, 1, 1, table_index, 0))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        self.assertTrue(self._model.moveRows(table_index, 0, 1, table_index, 2))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")

    def test_internal_move_mapping_row_by_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second"])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 2, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        mime_data = self._model.mimeData([self._model.index(1, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 0, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")

    def test_internal_move_first_mapping_item_to_second_by_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second", "third"])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 2, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "third")

    def test_internal_move_last_mapping_item_into_the_middle_by_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second", "third"])
        mime_data = self._model.mimeData([self._model.index(2, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 1, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "second")

    def test_internal_move_mapping_items_in_the_middle_by_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second", "third", "fourth"])
        mime_data = self._model.mimeData([self._model.index(1, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 3, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "fourth")
        mime_data = self._model.mimeData([self._model.index(2, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 1, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "fourth")

    def test_internal_move_multiple_mapping_items_by_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second", "third", "fourth"])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index), self._model.index(1, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 4, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "fourth")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "second")
        mime_data = self._model.mimeData([self._model.index(2, 0, table_index), self._model.index(3, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 0, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "fourth")
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index), self._model.index(3, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 2, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "fourth")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "third")
        mime_data = self._model.mimeData([self._model.index(1, 0, table_index), self._model.index(2, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 3, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "fourth")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "third")
        mime_data = self._model.mimeData([self._model.index(1, 0, table_index), self._model.index(3, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 0, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "fourth")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "first")
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index), self._model.index(2, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 4, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "fourth")
        self.assertEqual(self._model.index(3, 0, table_index).data(), "second")

    def test_undo_mapping_list_mime_data_drop(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", ["second", "third", "fourth"])
        mime_data = self._model.mimeData([self._model.index(2, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 0, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "second")
        self._undo_stack.undo()
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "third")
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, 3, 0, table_index))
        self.assertEqual(self._model.index(0, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "third")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "first")
        self._undo_stack.undo()
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")
        self.assertEqual(self._model.index(1, 0, table_index).data(), "second")
        self.assertEqual(self._model.index(2, 0, table_index).data(), "third")

    def test_cannot_drop_mappings_on_same_source_table_item(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", [])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index)])
        self.assertFalse(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, -1, -1, table_index))

    def test_dropping_mappings_on_another_source_table_copies_the_mappings(self):
        table_index_1 = self._append_source_table_with_mappings("source_table", "one", [])
        table_index_2 = self._append_source_table_with_mappings("destination_table", "two", [])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index_1)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, -1, -1, table_index_2))
        self.assertEqual(self._model.rowCount(table_index_1), 1)
        self.assertEqual(self._model.index(0, 0, table_index_1).data(), "one")
        self.assertEqual(self._model.rowCount(table_index_2), 2)
        self.assertEqual(self._model.index(0, 0, table_index_2).data(), "two")
        self.assertEqual(self._model.index(1, 0, table_index_2).data(), "one")

    def test_undoing_mapping_drop_on_another_source_table(self):
        table_index_1 = self._append_source_table_with_mappings("source_table", "one", [])
        table_index_2 = self._append_source_table_with_mappings("destination_table", "two", [])
        mime_data = self._model.mimeData([self._model.index(0, 0, table_index_1)])
        self.assertTrue(self._model.dropMimeData(mime_data, Qt.DropAction.CopyAction, -1, -1, table_index_2))
        self.assertEqual(self._model.rowCount(table_index_1), 1)
        self.assertEqual(self._model.index(0, 0, table_index_1).data(), "one")
        self.assertEqual(self._model.rowCount(table_index_2), 2)
        self.assertEqual(self._model.index(0, 0, table_index_2).data(), "two")
        self.assertEqual(self._model.index(1, 0, table_index_2).data(), "one")
        self._undo_stack.undo()
        self.assertEqual(self._model.rowCount(table_index_2), 1)
        self.assertEqual(self._model.index(0, 0, table_index_2).data(), "two")

    def test_undoing_delete_mapping_command(self):
        table_index = self._append_source_table_with_mappings("my_table", "first", [])
        self._undo_stack.push(DeleteMapping(table_index.row(), self._model, 0))
        self.assertEqual(self._model.rowCount(table_index), 0)
        self._undo_stack.undo()
        self.assertEqual(self._model.rowCount(table_index), 1)
        self.assertEqual(self._model.index(0, 0, table_index).data(), "first")


class TestMappingComponentsTable(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._model_parent = QObject()
        self._undo_stack = QUndoStack(self._model_parent)
        self._model = MappingsModel(self._undo_stack, self._model_parent)
        self._model.add_empty_row()
        self._table_index = self._model.index(1, 0)
        with signal_waiter(self._model.dataChanged) as waiter:
            self.assertTrue(self._model.setData(self._table_index, "table"))
            waiter.wait()
        self.assertTrue(self._model.insertRow(0, self._table_index))
        self._list_index = self._model.index(0, 0, self._table_index)

    def tearDown(self):
        self._model_parent.deleteLater()
        QApplication.processEvents()

    def test_data_when_mapping_object_class_without_objects_or_parameters(self):
        root_mapping = import_mapping_from_dict({"map_type": "ObjectClass", "name": None, "object": None})
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), root_mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 3)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Entity names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(0, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(1, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(2, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(0, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(1, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(2, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        self.assertEqual(self._model.index(0, 3, self._list_index).data(), "")
        self.assertEqual(self._model.index(1, 3, self._list_index).data(), "")
        self.assertEqual(self._model.index(2, 3, self._list_index).data(), "")

    def test_data_when_mapping_invalid_object_class_with_parameters(self):
        indexed_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": {"map_type": None},
            "parameter_type": "map",
            "value": {"map_type": None},
            "extra_dimensions": [{"map_type": None}],
        }
        mapping_dict = {
            "map_type": "ObjectClass",
            "name": None,
            "objects": None,
            "parameters": indexed_parameter_mapping_dict,
        }
        self._model.set_root_mapping(
            self._table_index.row(), self._list_index.row(), import_mapping_from_dict(mapping_dict)
        )
        self.assertEqual(self._model.rowCount(self._list_index), 9)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Entity names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes")
        index = self._model.index(8, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        for row in range(9):
            index = self._model.index(row, 1, self._list_index)
            self.assertEqual(index.data(), "None")
            index = self._model.index(row, 2, self._list_index)
            self.assertEqual(index.data(), None)
            self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
            self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_valid_object_class_with_pivoted_parameters(self):
        array_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": 2,
            "parameter_type": "array",
            "value": {"map_type": "row", "reference": 0},
        }
        mapping_dict = {"map_type": "ObjectClass", "name": 0, "objects": 1, "parameters": array_parameter_mapping_dict}
        self._model.set_root_mapping(
            self._table_index.row(), self._list_index.row(), import_mapping_from_dict(mapping_dict)
        )
        self.assertEqual(self._model.rowCount(self._list_index), 8)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Entity names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        index = self._model.index(0, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(1, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(2, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(3, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(4, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(5, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(6, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(7, 1, self._list_index)
        self.assertEqual(index.data(), "Row")
        index = self._model.index(0, 2, self._list_index)
        self.assertEqual(index.data(), 0 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(1, 2, self._list_index)
        self.assertEqual(index.data(), 1 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(2, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(3, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(4, 2, self._list_index)
        self.assertEqual(index.data(), 2 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(5, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(6, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(7, 2, self._list_index)
        self.assertEqual(index.data(), 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        for row in range(8):
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_valid_object_class_with_parameters(self):
        indexed_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": {"map_type": "column", "reference": 99},
            "parameter_type": "map",
            "value": {"reference": 23, "map_type": "column"},
            "extra_dimensions": [{"reference": "fifth column", "map_type": "column"}],
        }
        mapping_dict = {
            "map_type": "ObjectClass",
            "parameters": indexed_parameter_mapping_dict,
            "name": {"reference": "class_name", "map_type": "constant"},
            "objects": {"reference": "object_name", "map_type": "constant"},
        }
        mapping = import_mapping_from_dict(mapping_dict)
        table_name = "source table"
        header = ["1", "2", "3", "4", "fifth column"]
        mapping.polish(table_name, header, "")
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 9)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Entity names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes")
        index = self._model.index(8, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        index = self._model.index(0, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(1, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(2, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(3, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(4, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(5, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(6, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(7, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(8, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(0, 2, self._list_index)
        self.assertEqual(index.data(), "class_name")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(1, 2, self._list_index)
        self.assertEqual(index.data(), "object_name")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(2, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(3, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(4, 2, self._list_index)
        self.assertEqual(index.data(), 99 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(5, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(6, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(7, 2, self._list_index)
        self.assertEqual(index.data(), 5)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(8, 2, self._list_index)
        self.assertEqual(index.data(), 23 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        for row in range(9):
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_valid_object_class_with_nested_map(self):
        indexed_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": {"map_type": "column", "reference": 99},
            "parameter_type": "map",
            "value": {"reference": 23, "map_type": "column"},
            "extra_dimensions": [
                {"reference": "fifth column", "map_type": "column"},
                {"reference": "sixth column", "map_type": "column"},
            ],
        }
        mapping_dict = {
            "map_type": "ObjectClass",
            "parameters": indexed_parameter_mapping_dict,
            "name": {"reference": "class_name", "map_type": "constant"},
            "objects": {"reference": "object_name", "map_type": "constant"},
        }
        mapping = import_mapping_from_dict(mapping_dict)
        table_name = "source table"
        header = ["1", "2", "3", "4", "fifth column", "sixth column"]
        mapping.polish(table_name, header, "")
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 11)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Entity names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names 1")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes 1")
        index = self._model.index(8, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names 2")
        index = self._model.index(9, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes 2")
        index = self._model.index(10, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        index = self._model.index(0, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(1, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(2, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(3, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(4, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(5, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(6, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(7, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(8, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(9, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(10, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(0, 2, self._list_index)
        self.assertEqual(index.data(), "class_name")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(1, 2, self._list_index)
        self.assertEqual(index.data(), "object_name")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(2, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(3, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(4, 2, self._list_index)
        self.assertEqual(index.data(), 99 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(5, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(6, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(7, 2, self._list_index)
        self.assertEqual(index.data(), 5)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(8, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(9, 2, self._list_index)
        self.assertEqual(index.data(), 6)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(10, 2, self._list_index)
        self.assertEqual(index.data(), 23 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        for row in range(11):
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_relationship_class_without_objects_or_parameters(self):
        mapping = import_mapping_from_dict(
            {"map_type": "RelationshipClass", "name": None, "object_classes": None, "object": None}
        )
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 4)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Dimension names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Element names")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        for row in range(4):
            index = self._model.index(row, 1, self._list_index)
            self.assertEqual(index.data(), "None")
            index = self._model.index(row, 2, self._list_index)
            self.assertEqual(index.data(), None)
            self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
            self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_invalid_relationship_class_with_parameters(self):
        indexed_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": {"map_type": None},
            "parameter_type": "map",
            "value": {"map_type": None},
            "extra_dimensions": [{"map_type": None}],
        }
        mapping_dict = {
            "map_type": "RelationshipClass",
            "name": None,
            "object_classes": None,
            "object": None,
            "parameters": indexed_parameter_mapping_dict,
        }
        self._model.set_root_mapping(
            self._table_index.row(), self._list_index.row(), import_mapping_from_dict(mapping_dict)
        )
        self.assertEqual(self._model.rowCount(self._list_index), 10)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Dimension names")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Element names")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names")
        index = self._model.index(8, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes")
        index = self._model.index(9, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        for row in range(9):
            index = self._model.index(row, 1, self._list_index)
            self.assertEqual(index.data(), "None")
            index = self._model.index(row, 2, self._list_index)
            self.assertEqual(index.data(), None)
            self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
            self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_multidimensional_relationship_class_with_parameters(self):
        indexed_parameter_mapping_dict = {
            "map_type": "parameter",
            "name": {"map_type": "column", "reference": 99},
            "parameter_type": "map",
            "value": {"reference": 23, "map_type": "column"},
            "extra_dimensions": [{"reference": "fifth column", "map_type": "column"}],
        }
        mapping_dict = {
            "map_type": "RelationshipClass",
            "name": {"map_type": "constant", "reference": "relationship_class name"},
            "object_classes": [
                {"map_type": "column", "reference": "column header"},
                {"map_type": "constant", "reference": "second class"},
            ],
            "objects": [{"map_type": "column", "reference": 21}, {"map_type": "column", "reference": 22}],
            "parameters": indexed_parameter_mapping_dict,
        }
        mapping = import_mapping_from_dict(mapping_dict)
        table_name = "source table"
        header = ["column header", "2", "3", "4", "fifth column", "sixth column"]
        mapping.polish(table_name, header, "")
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 12)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        index = self._model.index(0, 0, self._list_index)
        self.assertEqual(index.data(), "Entity class names")
        index = self._model.index(1, 0, self._list_index)
        self.assertEqual(index.data(), "Dimension names 1")
        index = self._model.index(2, 0, self._list_index)
        self.assertEqual(index.data(), "Dimension names 2")
        index = self._model.index(3, 0, self._list_index)
        self.assertEqual(index.data(), "Element names 1")
        index = self._model.index(4, 0, self._list_index)
        self.assertEqual(index.data(), "Element names 2")
        index = self._model.index(5, 0, self._list_index)
        self.assertEqual(index.data(), "Entity metadata")
        index = self._model.index(6, 0, self._list_index)
        self.assertEqual(index.data(), "Alternative names")
        index = self._model.index(7, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter names")
        index = self._model.index(8, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter value metadata")
        index = self._model.index(9, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter index names")
        index = self._model.index(10, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter indexes")
        index = self._model.index(11, 0, self._list_index)
        self.assertEqual(index.data(), "Parameter values")
        index = self._model.index(0, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(1, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(2, 1, self._list_index)
        self.assertEqual(index.data(), "Constant")
        index = self._model.index(3, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(4, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(5, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(6, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(7, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(8, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(9, 1, self._list_index)
        self.assertEqual(index.data(), "None")
        index = self._model.index(10, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(11, 1, self._list_index)
        self.assertEqual(index.data(), "Column")
        index = self._model.index(0, 2, self._list_index)
        self.assertEqual(index.data(), "relationship_class name")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(1, 2, self._list_index)
        self.assertEqual(index.data(), 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(2, 2, self._list_index)
        self.assertEqual(index.data(), "second class")
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(3, 2, self._list_index)
        self.assertEqual(index.data(), 21 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(4, 2, self._list_index)
        self.assertEqual(index.data(), 22 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(5, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(6, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(7, 2, self._list_index)
        self.assertEqual(index.data(), 99 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(8, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(9, 2, self._list_index)
        self.assertEqual(index.data(), None)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(10, 2, self._list_index)
        self.assertEqual(index.data(), 5)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        index = self._model.index(11, 2, self._list_index)
        self.assertEqual(index.data(), 23 + 1)
        self.assertEqual(index.data(Qt.ItemDataRole.BackgroundRole), None)
        self.assertFalse(index.data(Qt.ItemDataRole.ToolTipRole))
        for row in range(12):
            self.assertEqual(self._model.index(row, 3, self._list_index).data(), "")

    def test_data_when_mapping_object_class_with_regular_expression(self):
        root_mapping = import_mapping_from_dict({"map_type": "ObjectClass", "name": None, "object": None})
        root_mapping.filter_re = "starred"
        root_mapping.child.child.filter_re = "choose_me"
        self._model.set_root_mapping(self._table_index.row(), self._list_index.row(), root_mapping)
        self.assertEqual(self._model.rowCount(self._list_index), 3)
        self.assertEqual(self._model.columnCount(self._list_index), 4)
        self.assertEqual(self._model.index(0, 3, self._list_index).data(), "starred")
        self.assertEqual(self._model.index(1, 3, self._list_index).data(), "")
        self.assertEqual(self._model.index(2, 3, self._list_index).data(), "choose_me")


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``preview_updater`` module."""
from multiprocessing import Process, Queue
import pathlib
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication, QComboBox, QWidget
from spine_items.exporter.mvcmodels.full_url_list_model import FullUrlListModel
from spine_items.exporter.widgets.preview_updater import PreviewUpdater, WriteTableTask, write_task_loop
from spinedb_api import DatabaseMapping
from spinedb_api.export_mapping.export_mapping import AlternativeMapping
from spinedb_api.export_mapping.group_functions import NoGroup
from tests.mock_helpers import parent_widget


class TestPreviewUpdater(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._parent_widget = QWidget()

    def tearDown(self):
        self._parent_widget.deleteLater()

    def test_deleting_url_model_does_not_break_tear_down(self):
        with parent_widget() as window:
            setattr(window, "current_mapping_about_to_change", mock.MagicMock())
            setattr(window, "current_mapping_changed", mock.MagicMock())
            ui = mock.MagicMock()
            ui.database_url_combo_box = QComboBox(self._parent_widget)
            url_model = FullUrlListModel()
            mappings_table_model = mock.MagicMock()
            mappings_proxy_model = mock.MagicMock()
            mapping_editor_table_model = mock.MagicMock()
            project_dir = ""
            preview_updater = PreviewUpdater(
                window,
                ui,
                url_model,
                mappings_table_model,
                mappings_proxy_model,
                mapping_editor_table_model,
                project_dir,
            )
            # It is difficult to get the url_model deleted on C++ side.
            # We simulate it by emitting destroyed manually.
            url_model.destroyed.emit()
            self.assertIsNot(preview_updater._url_model, url_model)
            preview_updater.tear_down()


class TestWriteTaskLoop(unittest.TestCase):
    def test_quit(self):
        sender = Queue()
        receiver = Queue()
        sender.put("quit")
        write_task_loop(receiver, sender)
        self.assertTrue(receiver.get(), "finished")

    def test_quitting_takes_precedence_over_writing(self):
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(pathlib.Path(temp_dir) / "db.sqlite")
            with DatabaseMapping(url, create=True) as db_map:
                db_map.add_alternative_item(name="alt1")
                db_map.add_alternative_item(name="alt2")
                db_map.commit_session("Add test data.")
            db_map.close()
            sender = Queue()
            receiver = Queue()
            sender.put(WriteTableTask(url, "my mapping", 2.3, AlternativeMapping(0), True, 1, 3, NoGroup.NAME))
            sender.put("quit")
            write_task_loop(receiver, sender)
            self.assertTrue(receiver.get(), "finished")

    def test_writing(self):
        sender = Queue()
        receiver = Queue()
        process = Process(target=write_task_loop, args=(receiver, sender))
        process.start()
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(pathlib.Path(temp_dir) / "db.sqlite")
            with DatabaseMapping(url, create=True) as db_map:
                db_map.add_alternative_item(name="alt1")
                db_map.add_alternative_item(name="alt2")
                db_map.commit_session("Add test data.")
            db_map.close()
            sender.put(WriteTableTask(url, "my mapping", 2.3, AlternativeMapping(0), True, 1, 3, NoGroup.NAME))
            tables = receiver.get()
            self.assertEqual(tables, ((url, "my mapping"), "my mapping", {None: [["Base"], ["alt1"], ["alt2"]]}, 2.3))
            sender.put("quit")
            self.assertTrue(receiver.get(), "finished")
            process.join()

    def test_change_database_between_writes(self):
        sender = Queue()
        receiver = Queue()
        process = Process(target=write_task_loop, args=(receiver, sender))
        process.start()
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(pathlib.Path(temp_dir) / "db1.sqlite")
            with DatabaseMapping(url, create=True) as db_map:
                db_map.add_alternative_item(name="alt1")
                db_map.add_alternative_item(name="alt2")
                db_map.commit_session("Add test data.")
            db_map.close()
            sender.put(WriteTableTask(url, "my mapping", 2.3, AlternativeMapping(0), True, 1, 3, NoGroup.NAME))
            tables = receiver.get()
            self.assertEqual(tables, ((url, "my mapping"), "my mapping", {None: [["Base"], ["alt1"], ["alt2"]]}, 2.3))
            url = "sqlite:///" + str(pathlib.Path(temp_dir) / "db2.sqlite")
            with DatabaseMapping(url, create=True) as db_map:
                db_map.add_alternative_item(name="alt3")
                db_map.add_alternative_item(name="alt4")
                db_map.commit_session("Add test data.")
            db_map.close()
            sender.put(WriteTableTask(url, "my mapping", 23.0, AlternativeMapping(0), True, 1, 3, NoGroup.NAME))
            tables = receiver.get()
            self.assertEqual(tables, ((url, "my mapping"), "my mapping", {None: [["Base"], ["alt3"], ["alt4"]]}, 23.0))
            sender.put("quit")
            self.assertTrue(receiver.get(), "finished")
            process.join()


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Merger's execute kernel (do_work), as target for a multiprocess.Process"""
import os
from spine_engine.utils.helpers import create_log_file_timestamp
from spinedb_api.helpers import remove_credentials_from_url
from spinedb_api.spine_db_client import SpineDBClient


def do_work(process, cancel_on_error, logs_dir, from_server_urls, to_server_urls, lock, logger):
    from_clients = [SpineDBClient.from_server_url(server_url) for server_url in from_server_urls]
    from_url_export_data_response = [
        (from_client.get_db_url(), from_client.export_data()) for from_client in from_clients
    ]
    from_url_data = [
        (url, response["result"]) for url, response in from_url_export_data_response if response.get("result")
    ]
    all_errors = [response["error"] for _, response in from_url_export_data_response if "error" in response]
    for server_url in to_server_urls:
        to_client = SpineDBClient.from_server_url(server_url)
        with process.maybe_idle:
            to_client.db_checkin()
        for from_url, data in from_url_data:
            lock.acquire()
            try:
                response = to_client.import_data(data, "")
                if "error" in response:
                    all_errors.append(response["error"])
                    continue
                import_count, import_errors = response["result"]
                all_errors += import_errors
                if import_errors and cancel_on_error and import_count:
                    to_client.call_method("rollback_session")
                else:
                    sanitized_from_url = remove_credentials_from_url(from_url)
                    sanitized_to_url = remove_credentials_from_url(to_client.get_db_url())
                    if data:
                        result = to_client.call_method(
                            "commit_session", f"Import {import_count} items from {sanitized_from_url}"
                        )
                        if "error" in result:
                            all_errors.append(result["error"])
                        else:
                            logger.msg_success.emit(
                                f"Merged {import_count} items with {len(import_errors)} errors "
                                f"from {sanitized_from_url} into {sanitized_to_url}"
                            )
                    else:
                        logger.msg_warning.emit(f"No new data merged from {sanitized_from_url} into {sanitized_to_url}")
            finally:
                lock.release()
        to_client.db_checkout()
    if all_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_error.log"))
        with open(logfilepath, "w") as f:
            for err in all_errors:
                f.write(f"{err}\n")
        # Make error log file anchor with path as tooltip
        logfile_anchor = f"<a style='color:#BB99FF;' title='{logfilepath}' href='file:///{logfilepath}'>error log</a>"
        logger.msg_error.emit(f"Import errors. Logfile: {logfile_anchor}")
    return (True,)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Data connection properties widget."""
import os
from PySide6.QtCore import QPoint, Qt, QUrl, Slot
from spinetoolbox.widgets.properties_widget import PropertiesWidgetBase
from .custom_menus import DcDataContextMenu, DcRefContextMenu


class DataConnectionPropertiesWidget(PropertiesWidgetBase):
    """Widget for the Data Connection Item Properties."""

    def __init__(self, toolbox):
        """

        Args:
            toolbox (ToolboxUI): The toolbox instance where this widget should be embedded
        """
        from ..ui.data_connection_properties import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(toolbox)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        # Class attributes
        self.connect_signals()

    def connect_signals(self):
        """Connect signals to slots."""
        self.ui.treeView_dc_references.customContextMenuRequested.connect(self.show_references_context_menu)
        self.ui.treeView_dc_data.customContextMenuRequested.connect(self.show_data_context_menu)

    @Slot(QPoint)
    def show_references_context_menu(self, pos):
        """Create and show a context-menu in data connection properties
        references view.

        Args:
            pos (QPoint): Mouse position
        """
        index = self.ui.treeView_dc_references.indexAt(pos)
        dc = self._active_item
        global_pos = self.ui.treeView_dc_references.viewport().mapToGlobal(pos)
        dc_ref_context_menu = DcRefContextMenu(self, global_pos, index, dc)
        option = dc_ref_context_menu.get_action()
        dc_ref_context_menu.deleteLater()
        if option == "Open containing directory...":
            ref_path = self.ui.treeView_dc_references.model().itemFromIndex(index).data(Qt.ItemDataRole.DisplayRole)
            ref_dir = os.path.split(ref_path)[0]
            file_url = "file:///" + ref_dir
            self._toolbox.open_anchor(QUrl(file_url, QUrl.TolerantMode))
        elif option == "Open...":
            dc.open_reference(index)
        elif option == "Add file reference(s)...":
            dc.show_add_file_references_dialog()
        elif option == "Add database reference...":
            dc.show_add_db_reference_dialog()
        elif option == "Remove reference(s)":
            dc.remove_references()
        elif option == "Copy file reference(s) to project":
            dc.copy_to_project()
        elif option == "Refresh reference(s)":
            dc.refresh_references()
        elif option != "None":
            raise RuntimeError(f"Unknown menu option '{option}'")

    @Slot(QPoint)
    def show_data_context_menu(self, pos):
        """Create and show a context-menu in data connection properties
        data view.

        Args:
            pos (QPoint): Mouse position
        """
        index = self.ui.treeView_dc_data.indexAt(pos)
        dc = self._active_item
        global_pos = self.ui.treeView_dc_data.viewport().mapToGlobal(pos)
        dc_data_context_menu = DcDataContextMenu(self, global_pos, index, dc)
        option = dc_data_context_menu.get_action()
        dc_data_context_menu.deleteLater()
        if option == "New file...":
            dc.make_new_file()
        elif option == "Open...":
            dc.open_data_file(index)
        elif option == "Remove file(s)":
            dc.remove_files()
        elif option == "Open directory...":
            dc.open_directory()
        elif option != "None":
            raise RuntimeError(f"Unknown menu option '{option}'")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`ValueTransformTableModel`."""
from enum import IntEnum, unique
from PySide6.QtCore import QModelIndex, Qt
from ..commands import SetData
from .parameter_drop_target_table_model import ParameterDropTargetTableModel


@unique
class TransformationsTableColumn(IntEnum):
    CLASS = 0
    PARAMETER = 1
    INSTRUCTIONS = 2


@unique
class TransformationsTableRole(IntEnum):
    INSTRUCTIONS = Qt.ItemDataRole.UserRole + 1
    SILENT_EDIT = Qt.ItemDataRole.UserRole + 2


class ValueTransformationsTableModel(ParameterDropTargetTableModel):
    """A table model that contains parameter value transformations."""

    GET_DATA_ROLES = (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.DisplayRole, TransformationsTableRole.INSTRUCTIONS)
    SET_DATA_ROLES = (
        TransformationsTableRole.SILENT_EDIT,
        TransformationsTableRole.SILENT_EDIT,
        TransformationsTableRole.INSTRUCTIONS,
    )

    def __init__(self, settings, undo_stack, parent):
        """
        Args:
            settings (ValueTransformSettings, optional): initial settings
            undo_stack (QUndoStack): undo stack
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._undo_stack = undo_stack
        if settings is not None:
            self._instructions = [
                [klass, param, instr]
                for klass, param_instr in settings.instructions.items()
                for param, instr in param_instr.items()
            ]
        else:
            self._instructions = []

    def columnCount(self, parent=QModelIndex()):
        return 3

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            column = index.column()
            if column != TransformationsTableColumn.INSTRUCTIONS:
                return self._instructions[index.row()][column]
            instructions = self._instructions[index.row()][column]
            return ", ".join(t["operation"] for t in instructions)
        if role == TransformationsTableRole.INSTRUCTIONS:
            return self._instructions[index.row()][TransformationsTableColumn.INSTRUCTIONS]
        return None

    def flags(self, index):
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDropEnabled
        if index.column() != TransformationsTableColumn.INSTRUCTIONS:
            flags |= Qt.ItemIsEditable
        return flags

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return ("Class", "Parameter", "Instructions")[section]
        return None

    def insertRows(self, row, count, parent=QModelIndex()):
        rows = count * [["", "", []]]
        self.beginInsertRows(parent, row, row + count - 1)
        self._instructions = self._instructions[:row] + rows + self._instructions[row:]
        self.endInsertRows()
        return True

    @staticmethod
    def _make_drop_row(entity_class, parameter):
        return [entity_class, parameter, []]

    def removeRows(self, row, count, parent=QModelIndex()):
        self.beginRemoveRows(parent, row, row + count - 1)
        self._instructions = self._instructions[:row] + self._instructions[row + count :]
        self.endRemoveRows()
        return True

    def rowCount(self, parent=QModelIndex()):
        return len(self._instructions)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if role == Qt.ItemDataRole.EditRole:
            if not value or not isinstance(value, str):
                return False
            column = index.column()
            message = "change class name" if column == TransformationsTableColumn.CLASS else "change parameter name"
            previous = self._instructions[index.row()][column]
            self._undo_stack.push(SetData(message, index, value, previous, TransformationsTableRole.SILENT_EDIT))
            return True
        if role == TransformationsTableRole.SILENT_EDIT:
            if not isinstance(value, str):
                return False
            self._instructions[index.row()][index.column()] = value
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
            return True
        if role == TransformationsTableRole.INSTRUCTIONS and index.column() == TransformationsTableColumn.INSTRUCTIONS:
            if not isinstance(value, list):
                return False
            row = index.row()
            self._instructions[row][index.column()] = value
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole, TransformationsTableRole.INSTRUCTIONS])
            return True
        return False
